import copy
import numpy as np
import cv2
from sklearn.metrics import confusion_matrix

def similarity(intersection, union):
    if union > 0:
        similarity = np.array(intersection/float(union)).item()
    elif intersection == 0 and union == 0:
        similarity = 1.0
    else:
        similarity = 0.0

    return similarity

def jaccard(annotation, segmentation, iTypes, typeSimilarity):
    intersections = 0
    unions = 0
    metrics = {}
    unique = np.unique(annotation)
    unique = np.append(unique,np.unique(segmentation))
    unique = np.unique(unique)

    #print(unique)

    for i in unique:
        if i in iTypes:
            metric = {}
            ta = np.zeros_like(annotation, dtype='bool')
            tb = np.zeros_like(segmentation, dtype='bool')

            ta[annotation == i] = 1 # Convert to binary mask
            aa = ta.sum()
            metric['annotation area'] = aa.item()

            tb[segmentation == i] = 1 # Convert to binary mask
            ab = tb.sum()
            metric['segmentation area'] = ab.item()

            andim = ta*tb # Logical AND
            intersection = andim.sum()
            metric['intersection'] = intersection.item()
            intersections += int(intersection)

            orim = ta + tb # Logical OR
            union = orim.sum()
            metric['union'] = union.item()
            unions += int(union)

            if i in typeSimilarity:
                typeSimilarity[i]['intersection'] += int(intersection)
                typeSimilarity[i]['union'] += int(union)

            metric['similarity'] = similarity(intersection, union)
            metrics[iTypes[i]['name']] = metric

    iou = {}
    iou['objects'] = metrics
    iou['image'] = similarity(intersections, unions)
         
    return iou, typeSimilarity, unique

def confusionmatrix(labels, segmentaiton, classes, total_confusion = None):

    confusion = confusion_matrix(labels.flatten(),segmentaiton.flatten(), labels=classes)

    if total_confusion is None:
        total_confusion = confusion
    else:
        total_confusion += confusion
        
    return confusion, total_confusion

def eval_records(fie, objTypes, records, resultsPath):

    results = []
    if not os.path.exists( resultsPath):
        os.makedirs(resultsPath)

    typeSimilarity = {}
    for objType in objTypes:
        typeSimilarity[objType['index']] = {'union':0, 'intersection':0} 

    for id, record in enumerate(records):
        im = cv2.imread(record['image'],cv2.IMREAD_GRAYSCALE)
        an = cv2.imread(record['annotation'],cv2.IMREAD_GRAYSCALE)


        outputs = fie(tf.expand_dims(tf.constant(im),-1))
        seg = np.squeeze(outputs['class_ids'].numpy())

        record['similarity'], typeSimilarity, unique = jaccard(an, seg, iTypes, typeSimilarity)       

        resultsImg = ed.AnnotationPlots(objTypes, cv2.cvtColor(im,cv2.COLOR_GRAY2RGB), an, seg, resultsPath, record)
        record['review'] = resultsImg

        print('similarity: {:.2f}, unique: {} image: {}'.format(record['similarity']['image'], unique, record['image']))
        results.append(record)

    return results, typeSimilarity

def ColorizeAnnotation(ann, lut):
    annrgb = [cv2.LUT(ann, lut[:, i]) for i in range(3)]
    annrgb = np.dstack(annrgb) 
    return annrgb

def MergeIman(img, ann, lut, mean=None, stDev = None):
    if mean is not None and stDev is not None:
        img = (img*stDev) + mean

    ann = ColorizeAnnotation(ann, lut)
    img = (img*ann).astype(np.uint8)
    return img

def MergeImAnSeg(img, ann, seg, lut, mean=None, stdev=None, ann_text='Label', seg_text='Segmentation'):

    font = cv2.FONT_HERSHEY_COMPLEX_SMALL
    iman = MergeIman(img, ann, lut, mean, stdev)
    imseg = MergeIman(img, seg, lut, mean, stdev)

    iman = cv2.putText(iman, ann_text,(10,25), font, 1,(255,255,255),1,cv2.LINE_AA)
    imseg = cv2.putText(imseg, seg_text,(10,25), font, 1,(255,255,255),1,cv2.LINE_AA)

    iman = cv2.hconcat([iman, imseg])
    return iman

class DatasetResults:
    # imgSave: save image to path defined in imgSave
    # imRecord: recorde image in results
    def __init__(self, class_dictionary, batch_size=1, imStatistics=False, imgSave=None, imRecord=False, task='segmentation'):
        # Prepare datasets for similarity computation
        self.class_dictionary = class_dictionary
        self.batch_size = batch_size
        self.imStatistics = imStatistics
        self.imgSave = imgSave
        self.imRecord = imRecord
        self.task = task

        # Process inputs for evaluation
        self.classSimilarity = {}
        self.objTypes = {}
        for objType in class_dictionary['objects']:
            if objType['trainId'] not in self.objTypes:
                self.objTypes[objType['trainId']] = copy.deepcopy(objType)
                # set name to category for objTypes and id to trainId
                self.objTypes[objType['trainId']]['name'] = objType['category']
                self.objTypes[objType['trainId']]['id'] = objType['trainId']

        for i in self.objTypes:
            self.classSimilarity[i]={'intersection':0, 'union':0}

        self.num_classes = len(self.objTypes)
        self.confusion_labels = range(self.num_classes)
 
        self.CreateLut()

        # Prepaire results data structures
        self.typeSimilarity = {}
        self.images = []
        self.totalConfusion=None
        self.dtSum =0

    def CreateLut(self):
        self.lut = np.zeros([256,3], dtype=np.uint8)
        for obj in self.class_dictionary ['objects']: # Load RGB colors as BGR
            self.lut[obj['trainId']][0] = obj['color'][2]
            self.lut[obj['trainId']][1] = obj['color'][1]
            self.lut[obj['trainId']][2] = obj['color'][0]
        self.lut = self.lut.astype(float) * 1/255. # scale colors 0-1
        self.lut[self.class_dictionary ['background']] = [1.0,1.0,1.0] # Pass Through

    def infer_results(self, iBatch, images, labels, segmentations, mean, stdev, dt):
        self.dtSum += dt
        numImages = len(images)
        dtImage = dt/numImages
        for j in range(numImages):

            result = {'dt':dtImage}

            if self.imgSave is not None or self.imRecord:
                imanseg = MergeImAnSeg(images[j], labels[j], segmentations[j], self.lut, mean[j], stdev[j])
                if self.imgSave is not None:
                    savename = '{}/{}{:04d}.png'.format(self.imgSave, self.task, self.batch_size*iBatch+j)
                    cv2.imwrite(savename, imanseg)
                if self.imRecord:
                    result['image'] = imanseg

            result['similarity'], self.classSimilarity, unique = jaccard(labels[j], segmentations[j], self.objTypes, self.classSimilarity)

            confusion = confusion_matrix(labels[j].flatten(),segmentations[j].flatten(), labels=self.confusion_labels)
            if self.totalConfusion is None:
                self.totalConfusion = confusion
            else:
                self.totalConfusion += confusion

            result['confusion'] = confusion.tolist()

            self.images.append(result)

        return self.totalConfusion

    def Results(self):

        dataset_similarity = {}

        num_images = len(self.images)
        average_time = self.dtSum/num_images
        sumIntersection = 0
        sumUnion = 0
        miou = 0
        for key in self.classSimilarity:
            intersection = self.classSimilarity[key]['intersection']
            sumIntersection += intersection
            union = self.classSimilarity[key]['union']
            sumUnion += union
            class_similarity = similarity(intersection, union)
            miou += class_similarity

            # convert to int from int64 for json.dumps
            dataset_similarity[key] = {'intersection':int(intersection) ,'union':int(union) , 'similarity':class_similarity}

        # miou computation
        positives = np.diagonal(self.totalConfusion)
        total = np.sum(self.totalConfusion,0)+np.sum(self.totalConfusion,1)-positives
        # Remoze zero values
        positives = positives[np.nonzero(total)]
        total = total[np.nonzero(total)]
        iou = positives/total
        miou =np.sum(iou)/self.num_classes

        results = {'confusion':self.totalConfusion.tolist(), 
                   'similarity':dataset_similarity,
                   'average time': average_time,
                   'mean intersection over union': miou,
                   'num images': num_images,
                }
        if self.imStatistics:
            results['images'] = self.images, 

        return results


import torch
import math
import numpy as np

def GaussianBasis(x, zero=0.0, sigma=0.33):
    return torch.exp(-1*torch.square((x-zero)/(2*sigma*sigma))) # torch.square not supported by torch.onnx

def NormGausBasis(len, i, depth, r=1.0):
        den = 0.0
        num = 0.0
        for j in range(len):
            bias = GaussianBasis(j,depth,r)
            if j==i:
                num=bias
            den = den + bias
        return num/den

# Exponential function from vertex and point
class Exponential():
    def __init__(self,vx=0.0, vy=0.0, px=1.0, py=1.0, power=2.0):
        self.vx = vx
        self.vy = vy
        self.px = px
        self.py = py
        if power < 0:
            raise ValueError('Exponential error power {} must be >= 0'.format(power))
        self.power = power
        if px <= vx:
            raise ValueError('Exponential error px={} must be > vx'.format(px, vx))
        else:
            self.a = (py-vy)/np.power(px-vx,self.power)
    def f(self, x):
        dx = x-self.vx
        y = self.a*np.power(x-self.vx,self.power) + self.vy
        return y


import torch
from prettytable import PrettyTable

def count_parameters(model):
    table = PrettyTable(["Modules", "Parameters"])
    total_params = 0
    for name, parameter in model.named_parameters():
        if not parameter.requires_grad: continue
        param = parameter.numel()
        table.add_row([name, param])
        total_params+=param
    print(table)
    print(f"Total Trainable Params: {total_params}")
    return total_params

def model_weights(model):
    total_params = 0
    for name, parameter in model.named_parameters():
        if not parameter.requires_grad: continue
        param = parameter.numel()
        total_params+=param
    return total_params

# compute network parameter stats
def model_stats(model):
    weight_array = None
    bias_array = None
    for name, parameter in model.named_parameters():
        if not parameter.requires_grad: continue
        param = parameter.numel()

        array = torch.flatten(parameter)

        if name.split('.')[-1] == 'weight':
            if weight_array is None:
                weight_array = array
            else:
                weight_array = torch.cat((weight_array, array))

        elif name.split('.')[-1] == 'bias':
            if bias_array is None:
                bias_array = array
            else:
                bias_array = torch.cat((bias_array, array))

    weight_std, weight_mean = torch.std_mean(weight_array, unbiased=False)
    bias_std, bias_mean = torch.std_mean(bias_array, unbiased=False)

    return weight_std, weight_mean, bias_std, bias_mean


import sys
import os
from pycocotools import mask
import numpy as np
import cv2
import json
from collections import defaultdict
import unittest
import torch

from pymlutil.functions import *



class Test(unittest.TestCase):
    def test_GaussianBasis(self):
        self.assertEqual(GaussianBasis(torch.tensor(0.0), zero=0.0, sigma=0.33), 1.0)


import sys
import os

from .s3 import *
from .jsonutil import *
from pymlutil.imutil import ImUtil, ImTransform

def parse_arguments():
    parser = argparse.ArgumentParser(description='Process arguments')

    parser.add_argument('-d', '--debug', action='store_true',help='Wait for debuggee attach')   
    parser.add_argument('-debug_port', type=int, default=3000, help='Debug port')
    parser.add_argument('-debug_address', type=str, default='0.0.0.0', help='Debug port')

    parser.add_argument('-credentials', type=str, default='creds.yaml', help='Credentials file.')


    parser.add_argument('--PutDir', '-p', action='store_true',help='Get coco dataset') 
    parser.add_argument('-src', type=str, default=None, help='path to source directory')
    parser.add_argument('-set', type=str, default='dataset', help='set defined in credentials file')
    parser.add_argument('-dest', type=str, default=None, help='destindation in s3')

    parser.add_argument('-cocourl', type=json.loads, default=None, 
                        help='List of coco dataset URLs to load.  If none, coco 2017 datafiles are loaded from https://cocodataset.org/#download')

    args = parser.parse_args()
    return args

def main(args):

    s3, creds, s3def = Connect(args.credentials)

    if args.PutDir:
        if args.set not in s3def['sets']:
            print('PutDir failed: args.set {} not found in credentials file'.format(args.set))
        elif args.src is None:
            print('PutDir failed: args.src is None')
        elif args.dest is None:
            print('PutDir failed: args.dest is None')
        else:
            dest = '{}/{}'.format(s3def['sets'][args.set]['prefix']/ args.dest)
            s3.PutDir(s3def['sets'][args.set]['bucket'], args.source, dest)


    print('pymluitil complete')

    
if __name__ == '__main__':
    import argparse
    args = parse_arguments()

    if args.debug:
        print("Wait for debugger attach on {}:{}".format(args.debug_address, args.debug_port))
        import debugpy

        debugpy.listen(address=(args.debug_address, args.debug_port)) # Pause the program until a remote debugger is attached
        debugpy.wait_for_client()  # Pause the program until a remote debugger is attached
        print("Debugger attached")

    main(args)


import os
import json
import yaml

def WriteDictJson(outdict, path):

    jsonStr = json.dumps(outdict, indent=2, sort_keys=False)
    with open(path,"w") as f:
        f.write(jsonStr)
    return True

def ReadDictJson(filepath):
    jsondict = None
    try:
        with open(filepath) as json_file:
            jsondict = json.load(json_file)
        if not jsondict:
            print('Failed to load {}'.format(filepath))
    except:
        print('Failed to load {}'.format(filepath))
    return jsondict

def Dict2Json(outdict):
    jsonStr = json.dumps(outdict, sort_keys=False, indent=4)      
    return jsonStr

def Json2Dict(json_file):
    jsondict = json.load(json_file)
    return jsondict

def ReadDictYaml(filepath):
    yamldict = {}
    try:
        with open(filepath) as yaml_file:
            yamldict = yaml.safe_load(yaml_file)
        if not yamldict:
            print('Failed to load {}'.format(filepath))
    except ValueError:
        print('Failed to load {} error {}'.format(filepath, ValueError))
    return yamldict

def WriteDictYaml(outdict, path):
    yamlStr = yaml.dump(outdict, indent=2, sort_keys=False)
    with open(path,"w") as f:
        f.write(yamlStr)
    return True

def ReadDict(filepath):
    ext = os.path.splitext(filepath)[1]
    if ext=='.yaml':
        readDict = ReadDictYaml(filepath)
    elif ext=='.json':
        readDict = ReadDictJson(filepath)
    else:
        readDict = None
    return readDict

def WriteDict(outdict, filepath):
    ext = os.path.splitext(filepath)[1]
    if ext=='.yaml':
        readDict = WriteDictYaml(outdict, filepath)
    elif ext=='.json':
        readDict = WriteDictJson(outdict, filepath)
    else:
        readDict = None
    return readDict

def str2bool(v):
    if isinstance(v, bool):
        return v
    if isinstance(v, int):
        return not(v==0)
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


import os
import io
import glob
import json, yaml
from datetime import datetime, timedelta
from pathlib import Path, PurePath
from tqdm import tqdm
import natsort as ns
from minio import Minio
import urllib3
import certifi

from .jsonutil import ReadDict

def remove_prefix(text, prefix):
    return text[text.startswith(prefix) and len(prefix):]

class s3store:

    def __init__(self, address, access_key, secret_key, tls = True, cert_verify=True, cert_path = None, timeout=5.0):
        self.addresss = address
        self.tls = tls
        urllib3.disable_warnings()
        if(cert_verify):
            if(cert_path):
                cert_location = cert_path
            else:
                cert_location = certifi.where()
            http = urllib3.PoolManager(cert_reqs='CERT_REQUIRED', ca_certs=cert_location, timeout=timeout, maxsize=10, block=True)
        else:
            http = urllib3.PoolManager(cert_reqs='CERT_NONE', timeout=timeout, maxsize=10, block=True)

        self.s3 = Minio(address,
             access_key=access_key,
             secret_key=secret_key,
             secure=tls,
             http_client = http)

    def GetUrl(self, bucket, path,  expires=timedelta(hours=2)):

        url = self.s3.presigned_get_object(bucket, path, expires=expires)
        return url

    def ListBuckets(self):
        return self.s3.list_buckets()

    def MakeBucket(self, bucket, location='us-east-1', object_lock=False):
        if not self.s3.bucket_exists(bucket):
            self.s3.make_bucket(bucket, location, object_lock)

    def PutFile(self, bucket, file, setname):

        success = True
        results = None

        try:
            self.MakeBucket(bucket)
            filename = setname+'/'+ os.path.basename(file)
            results = self.s3.fput_object(bucket, filename, file)
        except Exception as err:
            print(err)
            success = False

        return success

    def PutDir(self, bucket, path, setname):
        success = True
        files = list(Path(path).rglob('**/*.*'))

        try:
            self.MakeBucket(bucket)
            if path[len(path)-1] != '/':
                path = path +'/'
            for file in tqdm(files, total=len(files)):
                # Create object directory structure
                object_name = setname+ remove_prefix(str(file), str(Path(path)))
                self.s3.fput_object(bucket, object_name.lstrip('/'), str(file))
        except Exception as err:
            print(err)
            success = False

        return success

    def GetDir(self, bucket, setname, destdir):
        fileCount = 0

        # List all object paths in bucket that begin with my-prefixname.
        try:
            objects = self.s3.list_objects(bucket, prefix=setname, recursive=True)
            fileCount = len(tuple(objects))
            if(fileCount > 0):
                objects = self.s3.list_objects(bucket, prefix=setname, recursive=True) # Recreate object to reset iteratorator to beginning
                for obj in tqdm(objects, total=fileCount):
                    try:
                        
                        destination = '{}/{}'.format(destdir,remove_prefix(obj.object_name, setname))
                        if obj.is_dir:
                            if not os.path.isdir(destination):
                                os.mkdir(destination)
                        else:
                            self.s3.fget_object(bucket, obj.object_name, destination)
                    except:
                        print('Failed to copy {}/{} to {}'.format(bucket, obj.object_name, destination))


        except Exception as err:
            print(err)
            fileCount = 0
        except:
            print('Failed to read {}/{}'.format(bucket, setname))
            fileCount = 0

        return fileCount

    def GetFile(self, bucket, object_name, destination):
        success = True

        try:                     
            self.s3.fget_object(bucket, object_name, destination)
        except Exception as err:
            print(err)
            success = False
        except:
            print('Failed to copy {}/{} to {}'.format(bucket, object_name, destination))
            success = False

        return success

    def Mirror(self, bucket, setname, destdir):
        success = True

        try:
            objects = self.s3.list_objects(bucket, prefix=setname, recursive=True)
            fileCount = len(tuple(objects)) # Determine length for feedback during download
            if(fileCount > 0):
                objects = self.s3.list_objects(bucket, prefix=setname, recursive=True) # Recreate object to reset iteratorator to beginning
                for obj in tqdm(objects, total=fileCount):
                    try:
                        objstr = remove_prefix(obj.object_name, setname)
                        destination = '{}/{}'.format(destdir,objstr)
                        if obj.is_dir:
                            if not os.path.isdir(destination):
                                os.mkdir(destination)
                        else:
                            if not os.path.isfile(destination):
                                self.s3.fget_object(bucket, obj.object_name, destination)
                    except:
                        print('Failed to copy {}/{} to {}'.format(bucket, obj.object_name, destination))
                        success = False

        except Exception as err:
            print(err)
            raise err
        except:
            print('Failed to read {}/{}'.format(bucket, setname))
            success = False

        return success

    def ListObjects(self, bucket, setname=None, pattern='**', recursive=False):
        success = True
        files = []
        # List all object paths in bucket that begin with my-prefixname.
        try:
            objects = self.s3.list_objects(bucket, prefix=setname, recursive=recursive)
            if pattern is not None and len(pattern) > 0:
                for obj in objects:
                    if PurePath(obj.object_name).match(pattern):
                        files.append(obj.object_name)
            else:
                for obj in objects:
                    files.append(obj.object_name)
        except Exception as err:
            print(err)
            raise err

        return files

    def GetObjects(self, bucket, objectNames):
        objects = []
        for obj in objectNames:
            try:
                objects.append(GetObject(self, bucket, obj))
            except:
                print('error reading {}/{}'.format(bucket, obj))
        return objects

    def GetObject(self, bucket, object_name):
        success = True
        response = None
        data = None
        # List all object paths in bucket that begin with my-prefixname.
        try:
            response = self.s3.get_object(bucket, object_name)
            data = response.data
        except Exception as err:
                print(err)
        except:
                print('error reading {}/{}'.format(bucket, object_name))
        finally:
            if response:
                response.close()
                response.release_conn()

        return data

    def PutObject(self, bucket, object_name, obj):
        success = True

        # List all object paths in bucket that begin with my-prefixname.
        try:
            self.MakeBucket(bucket)
            if not isinstance(obj, io.BytesIO):
                objStream = io.BytesIO(obj)
                self.s3.put_object(bucket, object_name, objStream, length=len(objStream.getvalue()))
            else:
                obj.seek(0)
                self.s3.put_object(bucket, object_name, data=obj, length=len(obj.getvalue()))

        except Exception as err:
            print(err)
            success = False
        except:
                success = False

        return success

    def GetDict(self, bucket, object_name):
        success = True
        response = None
        data_dict = None
        # List all object paths in bucket that begin with my-prefixname.
        try:
            response = self.s3.get_object(bucket, object_name)

            if response.data:
                ext = os.path.splitext(object_name)[1]
                if ext=='.yaml':
                    data_dict = yaml.safe_load(response.data)
                elif ext=='.json':
                    data_dict = json.loads(response.data)



        except Exception as err:
            print('s3 Response error {}'.format(err))
        except json.JSONDecodeError as err:
            print('JSONDecodeError {}'.format(err))
        except:
            print('error reading {}/{}'.format(bucket, object_name))
        finally:
            if response:
                response.close()
                response.release_conn()

        return data_dict

    def PutDict(self, bucket, object_name, dict_data):
        success = True

        # List all object paths in bucket that begin with my-prefixname.
        try:
            self.MakeBucket(bucket)
            obj = json.dumps(dict_data, sort_keys=False, indent=4).encode()

            objStream = io.BytesIO(obj)
            self.s3.put_object(bucket, object_name, objStream, length=len(obj))
        except Exception as err:
            print(err)
            success = False
        except:
                success = False
        return success

    def ListModels(self, bucket='', model_type='' ):
        models = []
        sized_models = {}
        try:
            objects = self.s3.list_objects(bucket, prefix='', recursive=False)
            for obj in objects:
                splitname = obj.object_name.split('-')
                fields = len(splitname)
                if splitname[fields-1] == model_type+'/' or splitname[fields-2] == model_type : # remove saved models with specialized size
                    modelResults = self.s3.list_objects(bucket, prefix=splitname, recursive=False)       
                    models.append(obj.object_name)
                    if obj.object_name not in sized_models:
                        sized_models[obj.object_name] = []
                elif splitname[fields-3] == model_type:
                    targetstring = splitname[fields-2]+'-'
                    basemodel = obj.object_name.replace(splitname[fields-2]+'-','')
                    resolution = splitname[fields-2].split('x')
                    resolution = [int(resolution[0]),int(resolution[1])]
                    if(basemodel not in sized_models):
                        sized_models[basemodel] = []

                    model = {'name':obj.object_name,
                             'resolution':resolution
                    }
                    sized_models[basemodel].append(model)



        except:
            print('Failed to read objects')
        return models, sized_models

    def ListModelsWithResults(self, bucket, model_type='', result_name='results.json'):
        models_results = []
        models = self.ListModels(bucket=bucket, model_type=model_type)
        try:
            for model in models:
                objects = self.s3.list_objects(bucket, prefix=model, recursive=False)
                for obj in objects:
                    if obj.object_name == model+result_name:
                        models_results.append(model)
                        break

        except:
            print('Failed to read objects')
        return models_results
    
    def ListCheckpoints(self, bucket='training'):
        checkpoints = []
        try:
            objects = self.s3.list_objects('training', prefix='', recursive=False)
            for obj in objects:                              
                checkpoints.append(obj.object_name)

        except:
            print('Failed to read objects')
        return checkpoints
    def PutCheckpoint(self, bucket, checkpoint_path, checkpoint_name):
        success = True

        try:
            self.MakeBucket(bucket)
            ckpt = tf.train.latest_checkpoint(checkpoint_path)
            ckptfiles = glob.glob(ckpt+'.*')
            eventfiles = glob.glob(checkpoint_path+'/events.out.tfevents.*.*')
            eventfiles= ns.natsorted(eventfiles,alg=ns.PATH)

            # Add newest event file to lists
            ckptfiles.append(checkpoint_path+'/checkpoint')
            if len(eventfiles)>0:
                ckptfiles.append(eventfiles[-1])
            for ckfile in ckptfiles:
                obj = '{}/{}'.format(checkpoint_name,os.path.basename(ckfile))
                print('Write {} to {}/{}'.format(ckfile, bucket,obj))
                self.s3.fput_object(bucket, obj, ckfile)
        except Exception as err:
            print(err)
            raise err

        return success

    def GetCheckpoint(self, bucket, checkpoint_name, checkpoint_path):
        success = True
        # List all object paths in bucket that begin with my-prefixname.
        try:
            objects = self.s3.list_objects(bucket, prefix=checkpoint_name, 
                                            recursive=True)
            for obj in objects:
                print(obj.bucket_name, obj.object_name.encode('utf-8'), obj.last_modified,
                        obj.etag, obj.size, obj.content_type)

                        
                self.s3.fget_object(bucket, obj.object_name, '{}/{}'.format(checkpoint_path,os.path.basename(obj.object_name)))

        except Exception as err:
            print(err)
            raise err

        return checkpoints

    def CloneObjects(self, destbucket, destsetname, srcS3, srcbucket, srcsetname):
        success = True

        # List all object paths in bucket that begin with my-prefixname.
        try:
            objects = srcS3.list_objects(srcbucket, prefix=srcsetname, recursive=True)
            fileCount = len(tuple(objects)) 
            if(fileCount > 0):
                objects = srcS3.list_objects(srcbucket, prefix=setname, recursive=True)
                for obj in tqdm(objects, total=fileCount):
                    try:
                        destination = '{}/{}'.format(destsetname,remove_prefix(obj.object_name,srcsetname))

                        srcobj = srcS3.GetObject(srcbucket, obj.object_name)
                        self.PutObject(destbucket, destination, srcobj)

                        #if obj.is_dir:
                        #    if not os.path.isdir(destination):
                        #        os.mkdir(destination)
                        #else:
                        #    self.s3.fget_object(bucket, obj.object_name, destination)
                    except:
                        print('Failed to copy from {}/{} to {}/{}'.format(srcbucket,obj.object_name, destbucket, destination))
                        success = False            
            else:
                print('{}/{} contains {} objects'.format(srcbucket, setname, fileCount))

        except Exception as err:
            print(err)
            raise err

        return success

    def RemoveObjects(self, bucket, setname=None, pattern='**', recursive=False):
        success = True

        # List all object paths in bucket that begin with my-prefixname.
        # remove_objects minio API not working
        try:
            objects = self.s3.list_objects(bucket, prefix=setname, recursive=recursive)
            for obj in objects:
                self.s3.remove_object(bucket, obj.object_name)

        except Exception as err:
            print(err)
            success = False

        return success

def Connect(credentials_filename='creds.json', s3_name='store'):

    creds = ReadDict(credentials_filename)
    if not creds:
        print('Failed to load credentials file {}. Exiting'.format(credentials_filename))
        return False

    s3_creds = next(filter(lambda d: d.get('name') == s3_name, creds['s3']), None)
    
    s3 = s3store(s3_creds['address'], 
                 s3_creds['access key'], 
                 s3_creds['secret key'],
                 tls=s3_creds['tls'],
                 cert_verify=s3_creds['cert verify'],
                 cert_path=s3_creds['cert path'],
                )
                
    buckets = s3.ListBuckets()
    if not (len(buckets) >= 0) :
        s3 = None
        print('Failed connect to {}'.format(s3_creds['address']))

    del s3_creds['access key']
    del s3_creds['secret key']
    del s3_creds['cert verify']
    del s3_creds['cert path']
    
    return s3, creds, s3_creds


import numpy as np
import cv2



class ImUtil():
    def __init__(self, dataset_desc, class_dictionary):
        self.dataset_desc = dataset_desc
        self.class_dictionary = class_dictionary



        self.CreateLut()

    def isGrayscale(self):
        if 'image_colorspace' in self.dataset_desc:
            color_str = self.dataset_desc['image_colorspace']
            if color_str.lower() == 'grayscale':
                return True
        return False

    def CreateLut(self):
        self.lut = np.zeros([256,3], dtype=np.uint8)
        for obj in self.class_dictionary ['objects']: # Load RGB colors as BGR
            self.lut[obj['trainId']][0] = obj['color'][2]
            self.lut[obj['trainId']][1] = obj['color'][1]
            self.lut[obj['trainId']][2] = obj['color'][0]
        self.lut = self.lut.astype(float) * 1/255. # scale colors 0-1
        self.lut[self.class_dictionary ['background']] = [1.0,1.0,1.0] # Pass Through

    # Display functions
    def ColorizeAnnotation(self, ann):
        annrgb = [cv2.LUT(ann, self.lut[:, i]) for i in range(3)]
        annrgb = np.dstack(annrgb) 
        return annrgb

    def MergeIman(self, img, ann, mean=None, stDev = None):
        if mean is not None and stDev is not None:
            img = (img*stDev) + mean

        if self.class_dictionary is not None:
            ann = self.ColorizeAnnotation(ann)

        if self.isGrayscale():
            img = cv2.cvtColor(img,cv2.COLOR_GRAY2RGB)

        img = (img*ann).astype(np.uint8)
        return img

    def DisplayImAn(self, img, ann, seg, mean, stdev):

        font = cv2.FONT_HERSHEY_COMPLEX_SMALL
        iman = self.MergeIman(img, ann, mean, stdev)
        imseg = self.MergeIman(img, seg, mean, stdev)

        iman = cv2.putText(iman, 'Segmentation',(10,25), font, 1,(255,255,255),1,cv2.LINE_AA)
        imseg = cv2.putText(imseg, 'TensorRT',(10,25), font, 1,(255,255,255),1,cv2.LINE_AA)

        im = cv2.hconcat([iman, imseg])
        return im

class ImTransform():
    def __init__(self, 
        height=640, 
        width=640, 
        normalize=True, 
        enable_transform=True, 
        flipX=True, 
        flipY=False, 
        rotate=3, 
        scale_min=0.75, 
        scale_max=1.25, 
        offset=0.1,
        astype='float32',
        borderType=cv2.BORDER_CONSTANT,
        borderValue=0
    ):
        self.height = height
        self.width = width

        self.normalize = normalize
        self.enable_transform = enable_transform
        self.flipX = flipX
        self.flipY = flipY
        self.rotate = rotate
        self.scale_min = scale_min
        self.scale_max = scale_max
        self.offset = offset
        self.astype = astype
        self.borderType = borderType
        self.borderValue = borderValue



    # Expect img.shape[0]==ann.shape[0] and ann.shape[0]==ann.shape[0]
    def random_resize_crop_or_pad(self, img, ann):
        imgMean = None
        imgStd = None
        imgtype = img.dtype.name
        if self.normalize:
            imgMean = np.mean(img)
            imgStd = np.std(img)
            if imgStd > 0.0:
                img = (img - imgMean)/imgStd
            else:
                print('ImagesDataset.random_resize_crop_or_pad: imgStd is 0.0')
        
        if self.astype is not None:
            img = img.astype(self.astype)
        elif img.dtype.name is not  imgtype:
            img = img.astype(imgtype)

        height = img.shape[0]
        width = img.shape[1]
        
        # Pad
        pad = False
        top=0
        bottom=0
        left=0
        right=0
        if self.height > height:
            bottom = int((self.height-height)/2)
            top = self.height-height-bottom
            pad = True
        if self.width > width:
            right = int((self.width-width)/2)
            left = self.width-width-right
            pad = True

        if pad:
            img = cv2.copyMakeBorder(img, top, bottom, left, right, self.borderType, None, self.borderValue)
            ann = cv2.copyMakeBorder(ann, top, bottom, left, right, self.borderType, None, self.borderValue)

        # Transform
        if self.enable_transform:
                height, width = img.shape[:2]

                matFlip = np.identity(3)
                if self.flipX and np.random.choice(np.array([True, False])):
                    matFlip[0,0] *= -1.0
                    matFlip[0,2] += width-1
                if self.flipY and np.random.choice(np.array([True, False])):
                    matFlip[1,1] *= -1.0
                    matFlip[1,2] += height-1

                scale = np.random.uniform(self.scale_min, self.scale_max)
                angle = np.random.uniform(-self.rotate, self.rotate)
                offsetX = width*np.random.uniform(-self.offset, self.offset)
                offsetY = height*np.random.uniform(-self.offset, self.offset)
                center = (width/2.0 + offsetX, height/2.0 + offsetY)
                matRot = cv2.getRotationMatrix2D(center, angle, scale)
                matRot = np.append(matRot, [[0,0,1]],axis= 0)

                mat = np.matmul(matFlip, matRot)
                mat = mat[0:2]


                img = cv2.warpAffine(src=img, M=mat, dsize=(width, height))
                ann = cv2.warpAffine(src=ann, M=mat, dsize=(width, height))

        # Crop
        height = img.shape[0]
        width = img.shape[1]
        maxX = width - self.width
        maxY = height - self.height

        crop = False
        startX = 0
        startY = 0
        if maxX > 0:
            startX = np.random.randint(0, maxX)
            crop = True
        if  maxY > 0:
            startY = np.random.randint(0, maxY)
            crop = True
        if crop:

            img = img[startY:startY+self.height, startX:startX+self.width]
            ann = ann[startY:startY+self.height, startX:startX+self.width]

        return img, ann, imgMean, imgStd


#!/usr/bin/python3

import sys
import os
import requests
import random
import json
from datetime import datetime, timedelta

sys.path.append('.')
sys.path.append('./workflow')
from utils.s3 import s3store, Connect
from utils.jsonutil import ReadDict, Dict2Json

# paraemters is a dictionary of parameters to set
def set_parameters(workflow, new_parameters, template_name='train'):
    parameters = workflow['workflow']['spec']['arguments']['parameters']
    for parameter in parameters:
        for key, value in new_parameters.items():
            if key == parameter['name']:
                parameter['value'] = value

    '''
    templates = workflow['workflow']['spec']['templates']
    template = next(filter(lambda d: d.get('name') == template_name, templates), None)
    if template:
        parameters = template['inputs']['parameters']
        for parameter in parameters:
            name = parameter['name']
            if name in new_parameters:
                parameter['value'] = new_parameters[name]
    '''


def run(workflow, argocreds):
    session = requests.session()

    workflowstr = '{}://{}/api/v1/workflows/{}'.format(
        'https' if argocreds['tls'] else 'http',
        argocreds['address'],
        argocreds['namespace'])

    tasks_resp = session.post(workflowstr, json=workflow, verify = False)
    print('url: {} \nstatus_code: {} \nresponse: {}'.format(tasks_resp.url, tasks_resp.status_code, tasks_resp.text))
    return 0 if tasks_resp.ok == True else -1 


def parse_arguments():
    import argparse
    parser = argparse.ArgumentParser(description='Process arguments')

    parser.add_argument('--debug', '-d', action='store_true',help='Wait for debuggee attach')   
    parser.add_argument('-debug_port', type=int, default=3000, help='Debug port')
    parser.add_argument('-debug_address', type=str, default='0.0.0.0', help='Debug port')
    parser.add_argument('-test', action='store_true', help='Run unit tests')

    parser.add_argument('-credentails', type=str, default='creds.json', help='Credentials file.')
    parser.add_argument('-objectserver', type=str, default='store', help='Object server name.')

    parser.add_argument('--server', '-s', type=str, default=None, help='Argo Server.')

    parser.add_argument('--run', '-r', type=str, default='workflow/train_dev.yaml', help='Run workflow')
    parser.add_argument('--params', '-p', type=json.loads, default=None, help='Parameteers parsed by set_parameters  e.g.: {"output_name":"20220330h", "target_structure":0.0, "batch_size":3}')

    args = parser.parse_args()
    return args

def Test(args):

    s3, creds, s3def = Connect(args.credentails, s3_name=args.objectserver)
    if not s3:
        print("Failed to connect to s3 {} name {} ".format(args.credentails, args.objectserver))
        return -1

    argocreds = None
    if 'argo' in creds:
        if args.server is not None:
            argocreds = next(filter(lambda d: d.get('name') == args.server, creds['argo']), None)
        else:
            #argocreds = random.choice(creds['argo'])
            argocreds = creds['argo'][0]

    if not argocreds:
        print("Failed to find argo credentials for {}".format(args.server))
        return -1

    workflow = ReadDict(args.run)
    if not workflow:
        print('Failed to read {}'.format(args.run))
        return -1
    if args.params is not None and len(args.params) > 0:
        set_parameters(workflow, args.params)
    result = run(workflow, argocreds)

    return result


if __name__ == '__main__':
    args = parse_arguments()

    if args.debug:
        print("Wait for debugger attach")
        import debugpy
        debugpy.listen(address=(args.debug_address, args.debug_port))
        # Pause the program until a remote debugger is attached

        debugpy.wait_for_client()
        print("Debugger attached")

    result = Test(args)
    sys.exit(result)